@@ -2,7 +2,7 @@ package Proc::PersistentControl; # -*-perl-*-
 #
 # Author: Michael Staats 2014
 #
-# $Id: PersistentControl.pm 708 2015-02-24 16:37:03Z michael $
+# $Id: PersistentControl.pm 711 2015-03-11 14:29:23Z michael $
 #
 
 =head1 NAME
@@ -16,7 +16,8 @@ Proc::PersistentControl - Start and Control Background Processes ("jobs", proces
     my $controller = Proc::PersistentControl->new();
 
     my $Proc =
-        $controller->StartProc({ [ timeout => 42, Key => 'Value', ... ] },
+        $controller->StartProc({ [ timeout => 42, psynctime => 5,
+                                   Key => 'Value', ... ] },
                                "runme args");
 
     my @Procs = $controller->ProcList(['Key', 'Value']);
@@ -37,7 +38,9 @@ Proc::PersistentControl - Start and Control Background Processes ("jobs", proces
 
 WARNING: This module (and its pod) is beta.
 
-Work in progress. The interface might change. Probably there are bugs.
+The method that is used by this module is calling for race conditions all
+over the place... Since version 1.0 most of these should be found. But there
+will still be bugs.    
 
 This module creates background processes and allows to track
 them from multiple invocations of a perl program, i. e. not only
@@ -225,7 +228,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
 
 @ISA = qw(Exporter);
 
-$VERSION = '0.99';
+$VERSION = '1.0';
 
 use strict;
 use File::Path qw(mkpath rmtree); # legacy interface, works with older perls
@@ -241,23 +244,15 @@ BEGIN {
     }
 }
 
-######################################################################
-######################################################################
-
-my $debug = 0;
-
 ######################################################################
 # the new() method for the controller object
 #
+my $_debugfh = undef; # quick and dirty global var...
+
 sub new {
     my $class   = shift;
     my %opthash = @_;
 
-    if ($opthash{debug}) {
-	$opthash{verbose} = 1;
-	$debug = 1;
-    }
-
     $opthash{nannyname}  ||= 'ppc-nanny';
     $opthash{closefdmax} ||= 12;
     
@@ -280,6 +275,21 @@ sub new {
     croak "'$sdir' is not a directory and could not mkpath() it: $!\n"
 	if (not -d $sdir);
 
+    if ($opthash{debug}) {
+	my $file = $sdir . '/debug.out';
+	my $size = (stat($file))[7];
+	if ($size and $size > 102400) {
+	    rename($file, $file . '.old');
+	}
+	open($_debugfh, '>>', $file) ||
+	    carp "Cannot append to $file: $!";
+	select((select($_debugfh), $|=1)[0]);
+	eval { require Time::HiRes; };
+	if ($@) {
+	    print $_debugfh "No Time::HiRes ($@), sorry...\n";
+	}
+    }
+    
     my $rdir = $sdir . '/reaped';
     if (-e $rdir and not -d $rdir) {
 	carp "'$rdir' exists and is not a directory...\n";
@@ -289,39 +299,51 @@ sub new {
 	mkpath($rdir) or croak "Could not mkpath('$rdir'): $!\n";
     }
     
-    warn "new: blessing \$self with $class" if ($debug);
     bless $self, $class;
     return $self;
 }
 
+sub debug {
+    return undef unless $_debugfh;
+    my $clf = (caller(1))[3];
+    $clf ||=  (caller())[1];
+
+    my $msg = join(' ', @_);
+    $msg .= "\n" unless ($msg =~ /\n$/);
+    my $ts = $INC{'Time/HiRes.pm'} ? Time::HiRes::time() : time();
+    print $_debugfh $ts, ' ', $$, ' ', $clf, ': ', $msg;
+    return 1;
+}
+
 ######################################################################
 # system dependent stuff
 
 sub _pid_alive($) {
     my $pid  = shift;
-    
-    warn "_pid_alive: $pid" if ($debug);
+
+    debug($pid);
+
     return 0 if (not $pid or $pid <= 0);
     
     if ($^O !~ /^MSWin/) {
 	my $ret = kill(0, $pid);
-	warn "_pid_alive: returns $ret" if ($debug);
+	debug("returning $ret");
 	return $ret
     }
     
     # Windows
     my $process;
     if (Win32::Process::Open($process, $pid, 0)) {
-	warn "_pid_alive: Win32::Process::Open ok" if ($debug);
+	debug("Win32::Process::Open ok");
 	my $RC;
 	$process->GetExitCode($RC);
-	warn "_pid_alive: Win32 GetExitCode RC=$RC" if ($debug);
+	debug("Win32 GetExitCode RC=$RC");
 	if ($RC == Win32::Process::STILL_ACTIVE()) {
-	    warn "_pid_alive: returns 1" if ($debug);
+	    debug("returning 1");
 	    return 1;
 	}
     }
-    warn "_pid_alive: returns 0" if ($debug);
+    debug("returning 0");
     return 0;
 }
 
@@ -350,13 +372,16 @@ sub _unix_chld_sighandler {
     my $child;
     while (($child = waitpid(-1, WNOHANG)) > 0) {
 	my $RC = $? >> 8;
-	warn "_unix_chld_sighandler:$$:Child $child died with RC=$RC"
-	    if ($debug);
+	debug("Child $child died with RC=$RC (my grandchild $_unix_grandch_pid)");
 	if ($child == $_unix_grandch_pid) {
 	    # this is what we have been waiting for
 	    $_unix_grandch_pid = undef;
-	    open(RC, '>', $_unix_grandch_dir . "/RC=$RC"); close(RC);
-	    open(I, '>>', $_unix_grandch_dir . "/info");
+	    debug("Writing rc and endtime to $_unix_grandch_dir");
+	    open(RC, '>', $_unix_grandch_dir . "/RC=$RC") or
+		carp "Could not write $_unix_grandch_dir/RC=$RC: $!";
+	    close(RC);
+	    open(I, '>>', $_unix_grandch_dir . "/info") or
+		carp "Could not append to $_unix_grandch_dir/info: $!";
 	    print I "_endtime=", time(), "\n";
 	    close(I);
 	}
@@ -369,7 +394,7 @@ sub _unix_spawn {
     my $opt  = shift;
     my @cmd  = @_;
 
-    warn "_unix_spawn: $pdir\n" if ($debug);
+    debug("called with $pdir for $cmd[0] ...");
     
     $SIG{CHLD} = 'IGNORE'; # for now
     
@@ -377,7 +402,7 @@ sub _unix_spawn {
     croak "Cannot fork(): $!\n" if (not defined $childpid);
     
     if ($childpid == 0) {
-	warn "child: \$\$ = $$\n" if ($debug);
+	debug("I'm the new child for $opt->{nannyname} $cmd[0] ...");
 	# child
 	# fill the info directory,
 	    
@@ -402,7 +427,8 @@ sub _unix_spawn {
 
 	if ($opt->{_closefdmax} >= 3) {
 	    for (my $fd = 3; $fd <= $opt->{_closefdmax}; $fd++) {
-		POSIX::close($fd);
+		POSIX::close($fd)
+		    unless ($_debugfh and $fd == fileno($_debugfh));
 	    }
 	}
 	
@@ -414,10 +440,23 @@ sub _unix_spawn {
 	
 	$_unix_grandch_pid = fork(); # fork again for grandchild
 	if ($_unix_grandch_pid == 0) {
-	    warn "grandchild: \$\$ = $$\n" if ($debug);
+	    debug("I'm the new grandchild for $cmd[0] ...");
 	    #grandchild
 	    $SIG{CHLD} = 'DEFAULT';
 	    setpgrp(0, 0);
+	    # Sometimes (esp. on AIX) it seems that the following exec
+	    # can terminate before $_unix_grandch_pid is set in the
+	    # parent, which is bad since the sighandler checks it...
+	    # so wait for the pid file
+	    my $psynctime = $opt->{psynctime} || 10;
+	    my $waited = 0;
+	    debug("waiting max $psynctime s for 'start' file");
+	    for (my $nwait = 1; $nwait < 10 * $psynctime; $nwait++) {
+		last if (-e $pdir . "/start");
+		$waited++;
+		select(undef, undef, undef, 0.1);
+	    }
+	    debug("waited " . (0.1 * $waited) . " s for 'start' file");
 	    exec(@cmd) or
 		croak "Could not exec(", join(', ', @cmd), ")\n";
 	}
@@ -427,27 +466,32 @@ sub _unix_spawn {
 	$SIG{TERM} = \&_unix_term_sighandler;
 	$SIG{QUIT} = \&_unix_term_sighandler;
 
-	open(PID, '>', $pdir . "/pid=$_unix_grandch_pid"); close(PID);
-	
-	warn "child: \$_unix_grandch_pid = $_unix_grandch_pid\n" if ($debug);
+	open(PID, '>', $pdir . "/pid=$_unix_grandch_pid"); close(PID);	
+	open(STA, '>', $pdir . "/start"); print STA time(), "\n"; close(STA);
+	debug("child says: \$_unix_grandch_pid = $_unix_grandch_pid " .
+	      "for $cmd[0] ...");
 	
 	my $timeout = $opt->{timeout} ? int($opt->{timeout}) : 86400000;
 	my $maxtime = time + $timeout;
 
-	# $_unix_grandch_pid will be set to undef in the sighandler
-	while ($_unix_grandch_pid and time < $maxtime) {
+	# $_unix_grandch_pid will be set to undef in the sighandler...
+	# well, sometimes it seems it doesn't (some perls on some AIX, eg.)
+	# so we also test with kill(0, ...)
+	debug("Before wait loop for $cmd[0] ...");
+	while ($_unix_grandch_pid and kill(0, $_unix_grandch_pid) and time < $maxtime) {
 	    my $remain = $maxtime - time;
 	    # check each min to avoid race condition, see below
 	    $remain = 60 if ($remain > 60); 
-	    warn "$$: sleep($remain) at ", scalar(localtime), "\n" if ($debug);
+	    debug("Waiting for grandchild... sleep($remain) at "
+		  . scalar(localtime));
 	    sleep($remain) # will be interrupted by SIGCHLD, this is what we want
 		           # but if the sighandler jumps in exactly HERE
 		           # we still have a race cond.
 		           # That's why we sleep at most 60 s
 		if ($_unix_grandch_pid); # just to be sure...
 	}
-	warn "$$: after sleep loop at ", scalar(localtime), "\n" if ($debug);
-	if ($_unix_grandch_pid) {
+	debug("after wait loop for $cmd[0] ... at " . scalar(localtime));
+	if ($_unix_grandch_pid and kill(0, $_unix_grandch_pid)) {
 	    carp "ProcPersCtrl:$$: Child $_unix_grandch_pid is alive after $timeout s, killing it\n";
 	    open(I, '>>', $pdir . '/info') or
 		croak "Cannot append to '$pdir/info': $!\n";
@@ -458,7 +502,7 @@ sub _unix_spawn {
 	exit(0);
     }
     # parent
-    warn "parent: \$\$ = $$ child $childpid\n" if ($debug);
+    debug("parent says: child $childpid");
     return $childpid;
 }
 
@@ -474,13 +518,13 @@ sub _win_spawn {
 
     # construct args the DOS way, hopefully...
     my $Arg = 'perl ' . $helper . " ";
-    $Arg .= ' -d ' if ($debug);
+    $Arg .= ' -d ' if ($_debugfh);
     my $timeout = $opt->{timeout} ? $opt->{timeout} : 0;
     $Arg .= " -t $timeout " if ($timeout > 0);
     # well, if you have a '"' inside one of the arguments, I'm sorry...
     $Arg .= "\"$pdir\" \"" . join('" "', @cmd) . '"';
 
-    warn "_win_spawn: perl \$Arg = '$Arg'\n" if ($debug);
+    debug("perl \$Arg = '$Arg'");
 
     # launch windows process (helper script, in background)
     my $process;
@@ -489,7 +533,7 @@ sub _win_spawn {
 
     my $childpid = $process->GetProcessID();
     
-    warn "_win_spawn: \$childpid = $childpid" if ($debug);
+    debug("\$childpid = $childpid");
     
     open(I, '>', $pdir . '/info') or croak "Cannot write '$pdir/info': $!\n";
     my $start = time();
@@ -511,7 +555,7 @@ sub _unix_kill {
     my @pids = @_;
     my @ret;
     foreach my $p (@pids) {
-	warn "_unix_kill:$$:killing $p" if ($debug);
+	debug("killing $p");
 	# first the process group if we have one
 	kill(-15, $p) and
 	    (select(undef, undef, undef, 0.2), kill  -3, $p) and
@@ -529,7 +573,7 @@ sub _win_kill {
     my @pids = @_;
     foreach my $p (@pids) {
 	my $exitcode = 130;
-	warn "_win_kill: killing $p" if ($debug);
+	debug("killing $p");
 	Win32::Process::KillProcess($p, $exitcode);
     }
     return @pids;
@@ -561,11 +605,13 @@ sub _get_all_uids {
 
     opendir(D, $sdir) or croak "Cannot opendir '$sdir': $!\n";
     while (my $d = readdir(D)) {
-	warn "_get_all_uids: found \"$d\"" if ($debug);
-	push(@ret, $d) if ($d =~ m/^PPC-/);
+	if ($d =~ m/^PPC-/) {
+	    push(@ret, $d);
+	    debug("found job info dir \"$d\"");
+	}
     }
     closedir(D);
-    warn "_get_all_uids: returning ", join(' ', @ret) if ($debug);
+    debug("returning " . join(' ', @ret));
     return @ret;
 }
 
@@ -573,12 +619,12 @@ sub _get_pid_by_uid {
     my $sdir = shift;
     my $UID  = shift;
 
-    warn "_get_pid_by_uid: in $sdir $UID" if ($debug);
+    debug("in $sdir $UID");
     my $pdir = $sdir . '/' . $UID;
     my $pid;
     if (opendir(DD, $pdir)) {
 	while (my $f = readdir(DD)) {
-	    warn "_get_pid_by_uid: read $f" if ($debug);
+	    debug("read $f");
 	    $pid = $1 if ($f =~ m/pid=(\d+)$/);
 	}
 	closedir(DD);
@@ -586,7 +632,7 @@ sub _get_pid_by_uid {
     } else {
 	carp "Can't opendir $pdir for $UID\n";
     }
-    warn "_get_pid_by_uid: returns $pid" if ($debug);
+    debug("returning $pid");
     return ($pid);
 }
 
@@ -601,15 +647,27 @@ sub _get_info_by_uid {
     
     open(I, $infofile) or return undef;
 
+    while (<I>) {
+	chomp;
+	$ret{$1} = $2 if (m/([^=]+)=(.*)/);
+    }
+    close(I);
+    
     $ret{_dir} = $pdir;
     $ret{_dir} =~ s|/|\\|g if ($^O =~ /^MSWin/); # cosmetics, also works without
 
+    my $rctryt = $ret{psynctime} || 10;
+    
     my $pidalive = 0;
-    for (my $ntry = 0; $ntry < 10; $ntry++, select(undef, undef, undef, 0.1)) {
+    debug("waiting max $rctryt s for 'RC' file");
+    for (my $ntry = 0; $ntry < 10 * $rctryt; $ntry++, select(undef, undef, undef, 0.1)) {
 	# try a few times, just in case this function is called
 	# after the process has terminated but before RC could be written
+	# nevertheless it's still a race condition...
+	debug("try " . ($ntry + 1) . " to get RC in $pdir");
 	opendir(D, $pdir) or croak "cannot opendir $pdir: $!";
 	while ($_ = readdir(D)) {
+	    debug("readdir() gives: $_");
 	    $ret{_pid} = $1 if (m/^pid=(\d+)/);
 	    $ret{_RC}  = $1 if (m/^RC=(\d+)/);
 	}
@@ -619,7 +677,7 @@ sub _get_info_by_uid {
 	    $pidalive = _pid_alive($ret{_pid});
 	    last if ($pidalive); # still alive, also ok
 	}
-	warn "No RC found althoug process is dead..." if ($debug);
+	debug("No RC found in $pdir although process is dead...");
     }
     
     if (exists($ret{_RC})) {
@@ -630,18 +688,17 @@ sub _get_info_by_uid {
 	} else {
 	    # no RC and not alive => terminated the hard way...
 	    $ret{_alive} = 0;
+	    debug("Could not determine RC, setting it to 130");
+	    carp "Could not determine return code of job in $pdir, " .
+		"setting it to 130";
 	    $ret{_RC} = 130;
 	    open(RC, '>', $pdir . '/RC=130'); close(RC);
 	}
     }
-    
-    while (<I>) {
-	chomp;
-	$ret{$1} = $2 if (m/([^=]+)=(.*)/);
-    }
-    close(I);
-    warn "_get_info_by_uid: \$ret{_dir} = $ret{_dir} \$ret{_pid} = $ret{_pid} ",
-    "\$ret{_PPCUID} = $ret{_PPDUID} \$ret{TAG} = $ret{TAG}" if ($debug);
+        
+    no warnings "uninitialized";
+    debug("\$ret{_RC} = $ret{_RC} \$ret{_dir} = $ret{_dir} " .
+	  "\$ret{_pid} = $ret{_pid} \$ret{TAG} = $ret{TAG}");
 
     return \%ret;
 }
@@ -652,7 +709,7 @@ sub _make_uid_list {
     my @in = @_;
     my @out;
 
-    warn "_make_uid_list in: ", join('-', @in) if ($debug);
+    debug("in: ", join('-', @in));
 
     # empty input, return all UIDS
     if (not $in[0]) {
@@ -664,7 +721,7 @@ sub _make_uid_list {
 	    push(@out, $UID) if (-f $pdir . '/info');
 	}
     }
-    warn "_make_uid_list out: ", join('-', @out) if ($debug);
+    debug("out: ", join('-', @out));
     return @out;
 }
 
@@ -688,8 +745,7 @@ sub StartProc {
 	carp "$w start with '_'" if ($o =~ m/^_/);
     }
     
-    print "ProcPersCtrl: StartProc command '", join(' ', @cmd), "'\n"
-	if ($self->{_opts}->{verbose});	
+    debug("command '" . join(' ', @cmd) . "'");
 
     # create a directory for process information
     my $psd = tempdir('PPC-XXXX', DIR => $sdir);
@@ -710,9 +766,25 @@ sub StartProc {
 	_PPCUID => $UID,
 	_controller => $self
     };
-    bless $Proc, 'Proc::PersistentControl::Proc';
 
-    return $Proc;
+    my $psynctime = $opt->{psynctime} || 10;
+    my $waited = 0;
+    for (my $nwait = 1; $nwait < 10 * $psynctime; $nwait++) {
+	last if (-e $psd . "/start");
+	$waited++;
+	select(undef, undef, undef, 0.1);
+    }
+    debug("waited " . (0.1 * $waited) . " s for 'start' file");
+    if (-e $psd . "/start") {
+	bless $Proc, 'Proc::PersistentControl::Proc';
+	debug("returning Proc object for $UID, pid file " .
+	      (glob($psd . '/pid=*'))[0]);
+	return $Proc;
+    } else {
+	carp "StartProc could not start child process after " .
+	    (0.1 * $waited) . " s\n";
+	return undef;
+    }
 }
 
 sub ProcList {
@@ -724,6 +796,8 @@ sub ProcList {
     
     my @ret;
 
+    debug("called from " . join('-', caller()));
+    
     my @uidlist = _make_uid_list($self->{sdir});
     
     UID: foreach my $UID (@uidlist) {
@@ -747,11 +821,13 @@ sub RipeList {
     
     my $self       = shift;
     my @ret;
+
+    debug("called from " . join('-', caller()));
     
     my @uidlist = _make_uid_list($self->{sdir});
 
     foreach my $UID (@uidlist) {
-	warn "ProcPersCtrl: RipeList checking $UID\n" if ($debug);
+	debug("checking $UID");
 	if (not _UID_alive($self->{sdir}, $UID)) {
 	    my $Proc = {
 		_PPCUID => $UID,
@@ -759,8 +835,7 @@ sub RipeList {
 	    };
 	    bless $Proc, 'Proc::PersistentControl::Proc';
 	    push(@ret, $Proc);
-	    warn "ProcPersCtrl: $UID not alive, ready for reaping\n"
-		if ($debug);
+	    debug("$UID not alive, ready for reaping");
 	}
     }
     return(@ret);
@@ -781,6 +856,8 @@ sub _getUID {
 
 sub Kill {
     my $self = shift;
+
+    Proc::PersistentControl::debug("called from " . join('-', caller()));
     
     return undef unless (Proc::PersistentControl::_UID_alive(
 			     $self->{_controller}->{sdir},
@@ -797,7 +874,7 @@ sub Kill {
     # wait max 1 sec until process has finished and info is written
     for (my $nwait = 0; $nwait < 10; $nwait++) {
 	last if (not $self->IsAlive());
-	warn "Process not dead after Kill() ($nwait)..." if ($debug);
+	Proc::PersistentControl::debug("Process not dead after Kill() ($nwait)...");
 	select(undef, undef, undef, 0.1);
     }
     carp "Something is strange, Kill()ed process seems to be still alive..."
@@ -809,6 +886,8 @@ sub Kill {
 sub Info {
     my $self = shift;
 
+    Proc::PersistentControl::debug("called from " . join('-', caller()));
+    
     return Proc::PersistentControl::_get_info_by_uid(
 	$self->{_controller}->{sdir},
 	$self->{_PPCUID});
@@ -829,6 +908,9 @@ sub IsRipe {
 
 sub Reap {
     my $self = shift;
+
+    Proc::PersistentControl::debug("called from " . join('-', caller()));
+    
     return undef unless ($self->IsRipe());
 
     my $sdir = $self->{_controller}->{sdir};
@@ -836,12 +918,26 @@ sub Reap {
     
     my $source = $sdir . '/' . $UID;
     my $target = $sdir . '/reaped/' . $UID;
-    move($source, $target) or
-	# sometimes (on windows) the dir seems to be "locked"
-	# also after process termination etc.
-	(select(undef, undef, undef, 0.2), move($source, $target)) or
-	(select(undef, undef, undef, 0.8), move($source, $target)) or
-	        carp "Cannot move('$source', '$target'): $!\n";
+
+    if (not -d $target) {
+	rename($source, $target) or
+	move($source, $target) or
+	    # sometimes (on windows) the dir seems to be "locked"
+	    # also after process termination etc.
+	    (Proc::PersistentControl::debug("rename/1st move($source, $target) failed, trying again"),
+	     select(undef, undef, undef, 0.2),
+	     move($source, $target)) or
+	     (Proc::PersistentControl::debug("2nd move($source, $target) failed, trying again"),
+	      select(undef, undef, undef, 0.8),
+	      move($source, $target)) or
+	      (Proc::PersistentControl::debug("3rd move($source, $target) failed"));
+	# do not check return value of move
+	# this is unreliable, maybe (on windows...)
+	carp "Cannot move('$source', '$target'): $!\n" unless (-d $target);
+    } else {
+	### Reap called twice???
+	carp "Reap: '$target' already exists, Reap() called twice?\n"
+    }
     return Proc::PersistentControl::_get_info_by_uid($sdir . '/reaped', $UID);
 }
 
@@ -852,7 +948,6 @@ sub DESTROY {
     my $pdir = $self->{_controller}->{sdir} . '/reaped/' . $self->{_PPCUID};
 
     if (-d $pdir) {
-	warn "DESTROYing $pdir" if ($debug);
 	rmtree($pdir) or
 	    carp "Could not rmtree($pdir): $!";
     }
